@@ -1,8 +1,9 @@
 /* Directory management routines
 
-   Copyright (C) 1994, 1995, 1996 Free Software Foundation, Inc.
+   Copyright (C) 1994, 1995, 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2007
+     Free Software Foundation, Inc.
 
-   Converted for ext2fs by Miles Bader <miles@gnu.ai.mit.edu>
+   Converted for ext2fs by Miles Bader <miles@gnu.org>
 
    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
@@ -72,56 +73,83 @@ struct dirstat
 
   /* Index of this directory block. */
   int idx;
-  
-  /* For stat COMPRESS, this is the address (inside mapbuf) 
+
+  /* For stat COMPRESS, this is the address (inside mapbuf)
      of the first direct in the directory block to be compressed. */
   /* For stat HERE_TIS, SHRINK, and TAKE, this is the entry referenced. */
-  struct ext2_dir_entry *entry;
+  struct ext2_dir_entry_2 *entry;
 
   /* For stat HERE_TIS, type REMOVE, this is the address of the immediately
      previous direct in this directory block, or zero if this is the first. */
-  struct ext2_dir_entry *preventry;
+  struct ext2_dir_entry_2 *preventry;
 
   /* For stat COMPRESS, this is the number of bytes needed to be copied
      in order to undertake the compression. */
   size_t nbytes;
 };
 
-size_t diskfs_dirstat_size = sizeof (struct dirstat);
+const size_t diskfs_dirstat_size = sizeof (struct dirstat);
 
 /* Initialize DS such that diskfs_drop_dirstat will ignore it. */
-void 
+void
 diskfs_null_dirstat (struct dirstat *ds)
 {
   ds->type = LOOKUP;
 }
 
-static error_t					
-dirscanblock (vm_address_t blockoff, struct node *dp, int idx, char *name, 
-	      int namelen, enum lookup_type type, struct dirstat *ds,
-	      ino_t *inum);
+static error_t
+dirscanblock (vm_address_t blockoff, struct node *dp, int idx,
+	      const char *name, size_t namelen, enum lookup_type type,
+	      struct dirstat *ds, ino_t *inum);
 
+
+#if 0				/* XXX unused for now */
+static const unsigned char ext2_file_type[EXT2_FT_MAX] =
+{
+  [EXT2_FT_UNKNOWN]	= DT_UNKNOWN,
+  [EXT2_FT_REG_FILE]	= DT_REG,
+  [EXT2_FT_DIR]		= DT_DIR,
+  [EXT2_FT_CHRDEV]	= DT_CHR,
+  [EXT2_FT_BLKDEV]	= DT_BLK,
+  [EXT2_FT_FIFO]	= DT_FIFO,
+  [EXT2_FT_SOCK]	= DT_SOCK,
+  [EXT2_FT_SYMLINK]	= DT_LNK,
+};
+
+static const unsigned char file_type_ext2[] =
+{
+  [DT_UNKNOWN]	= EXT2_FT_UNKNOWN,
+  [DT_REG]	= EXT2_FT_REG_FILE,
+  [DT_DIR]	= EXT2_FT_DIR,
+  [DT_CHR]	= EXT2_FT_CHRDEV,
+  [DT_BLK]	= EXT2_FT_BLKDEV,
+  [DT_FIFO]	= EXT2_FT_FIFO,
+  [DT_SOCK]	= EXT2_FT_SOCK,
+  [DT_LNK]	= EXT2_FT_SYMLINK,
+};
+#endif
+
 /* Implement the diskfs_lookup from the diskfs library.  See
    <hurd/diskfs.h> for the interface specification.  */
 error_t
-diskfs_lookup_hard (struct node *dp, char *name, enum lookup_type type,
+diskfs_lookup_hard (struct node *dp, const char *name, enum lookup_type type,
 		    struct node **npp, struct dirstat *ds, struct protid *cred)
 {
   error_t err;
   ino_t inum;
-  int namelen;
+  size_t namelen;
   int spec_dotdot;
   struct node *np = 0;
-  int retry_dotdot = 0;
+  ino_t retry_dotdot = 0;
   vm_prot_t prot =
     (type == LOOKUP) ? VM_PROT_READ : (VM_PROT_READ | VM_PROT_WRITE);
   memory_object_t memobj;
   vm_address_t buf = 0;
   vm_size_t buflen = 0;
-  int blockaddr;
+  vm_address_t blockaddr;
   int idx, lastidx;
   int looped;
-  
+
   if ((type == REMOVE) || (type == RENAME))
     assert (npp);
 
@@ -130,12 +158,16 @@ diskfs_lookup_hard (struct node *dp, char *name, enum lookup_type type,
 
   spec_dotdot = type & SPEC_DOTDOT;
   type &= ~SPEC_DOTDOT;
-  
+
   namelen = strlen (name);
 
   if (namelen > EXT2_NAME_LEN)
-    return ENAMETOOLONG;
-  
+    {
+      if (ds)
+	diskfs_null_dirstat (ds);
+      return ENAMETOOLONG;
+    }
+
  try_again:
   if (ds)
     {
@@ -145,7 +177,7 @@ diskfs_lookup_hard (struct node *dp, char *name, enum lookup_type type,
     }
   if (buf)
     {
-      vm_deallocate (mach_task_self (), buf, buflen);
+      munmap ((caddr_t) buf, buflen);
       buf = 0;
     }
   if (ds && (type == CREATE || type == RENAME))
@@ -153,18 +185,23 @@ diskfs_lookup_hard (struct node *dp, char *name, enum lookup_type type,
 
   /* Map in the directory contents. */
   memobj = diskfs_get_filemap (dp, prot);
+
+  if (memobj == MACH_PORT_NULL)
+    return errno;
+
   buf = 0;
   /* We allow extra space in case we have to do an EXTEND. */
   buflen = round_page (dp->dn_stat.st_size + DIRBLKSIZ);
   err = vm_map (mach_task_self (),
 		&buf, buflen, 0, 1, memobj, 0, 0, prot, prot, 0);
   mach_port_deallocate (mach_task_self (), memobj);
+  if (err)
+    return err;
 
   inum = 0;
-  
-  if (!diskfs_check_readonly ())
-    dp->dn_set_atime = 1;
-  
+
+  diskfs_set_node_atime (dp);
+
   /* Start the lookup at DP->dn->dir_idx.  */
   idx = dp->dn->dir_idx;
   if (idx * DIRBLKSIZ > dp->dn_stat.st_size)
@@ -185,7 +222,7 @@ diskfs_lookup_hard (struct node *dp, char *name, enum lookup_type type,
 	}
       if (err != ENOENT)
 	{
-	  vm_deallocate (mach_task_self (), buf, buflen);
+	  munmap ((caddr_t) buf, buflen);
 	  return err;
 	}
 
@@ -200,8 +237,7 @@ diskfs_lookup_hard (struct node *dp, char *name, enum lookup_type type,
 	}
     }
 
-  if (!diskfs_check_readonly ())
-    dp->dn_set_atime = 1;
+  diskfs_set_node_atime (dp);
   if (diskfs_synchronous)
     diskfs_node_update (dp, 1);
 
@@ -225,7 +261,7 @@ diskfs_lookup_hard (struct node *dp, char *name, enum lookup_type type,
 		goto out;
 	    }
 	}
-	  
+
       /* We are looking up .. */
       /* Check to see if this is the root of the filesystem. */
       else if (dp->cache_id == 2)
@@ -233,7 +269,7 @@ diskfs_lookup_hard (struct node *dp, char *name, enum lookup_type type,
 	  err = EAGAIN;
 	  goto out;
 	}
-	  
+
       /* We can't just do diskfs_cached_lookup, because we would then deadlock.
 	 So we do this.  Ick.  */
       else if (retry_dotdot)
@@ -244,9 +280,9 @@ diskfs_lookup_hard (struct node *dp, char *name, enum lookup_type type,
 	      /* Drop what we *thought* was .. (but isn't any more) and
 		 try *again*. */
 	      diskfs_nput (np);
-	      mutex_unlock (&dp->lock);
+	      pthread_mutex_unlock (&dp->lock);
 	      err = diskfs_cached_lookup (inum, &np);
-	      mutex_lock (&dp->lock);
+	      pthread_mutex_lock (&dp->lock);
 	      if (err)
 		goto out;
 	      retry_dotdot = inum;
@@ -259,19 +295,19 @@ diskfs_lookup_hard (struct node *dp, char *name, enum lookup_type type,
 	  /* Lock them in the proper order, and then
 	     repeat the directory scan to see if this is still
 	     right.  */
-	  mutex_unlock (&dp->lock);
+	  pthread_mutex_unlock (&dp->lock);
 	  err = diskfs_cached_lookup (inum, &np);
-	  mutex_lock (&dp->lock);
+	  pthread_mutex_lock (&dp->lock);
 	  if (err)
 	    goto out;
 	  retry_dotdot = inum;
 	  goto try_again;
 	}
-      
+
       /* Here below are the spec dotdot cases. */
       else if (type == RENAME || type == REMOVE)
 	np = ifind (inum);
-      
+
       else if (type == LOOKUP)
 	{
 	  diskfs_nput (dp);
@@ -282,7 +318,7 @@ diskfs_lookup_hard (struct node *dp, char *name, enum lookup_type type,
       else
 	assert (0);
     }
-  
+
   if ((type == CREATE || type == RENAME) && !inum && ds && ds->stat == LOOKING)
     {
       /* We didn't find any room, so mark ds to extend the dir */
@@ -299,7 +335,7 @@ diskfs_lookup_hard (struct node *dp, char *name, enum lookup_type type,
       || !ds
       || ds->type == LOOKUP)
     {
-      vm_deallocate (mach_task_self (), buf, buflen);
+      munmap ((caddr_t) buf, buflen);
       if (ds)
 	ds->type = LOOKUP;	/* set to be ignored by drop_dirstat */
     }
@@ -308,7 +344,7 @@ diskfs_lookup_hard (struct node *dp, char *name, enum lookup_type type,
       ds->mapbuf = buf;
       ds->mapextent = buflen;
     }
-  
+
   if (np)
     {
       assert (npp);
@@ -342,20 +378,20 @@ diskfs_lookup_hard (struct node *dp, char *name, enum lookup_type type,
    diskfs_lookup.  If found, set *INUM to the inode number, else
    return ENOENT.  */
 static error_t
-dirscanblock (vm_address_t blockaddr, struct node *dp, int idx, char *name,
-	      int namelen, enum lookup_type type, 
+dirscanblock (vm_address_t blockaddr, struct node *dp, int idx,
+	      const char *name, size_t namelen, enum lookup_type type,
 	      struct dirstat *ds, ino_t *inum)
 {
-  int nfree = 0;
-  int needed = 0;
+  size_t nfree = 0;
+  size_t needed = 0;
   vm_address_t currentoff, prevoff;
-  struct ext2_dir_entry *entry = 0;
+  struct ext2_dir_entry_2 *entry = 0;
   int nentries = 0;
   size_t nbytes = 0;
   int looking = 0;
   int countcopies = 0;
   int consider_compress = 0;
-  
+
   if (ds && (ds->stat == LOOKING
 	     || ds->stat == COMPRESS))
     {
@@ -368,8 +404,8 @@ dirscanblock (vm_address_t blockaddr, struct node *dp, int idx, char *name,
        currentoff < blockaddr + DIRBLKSIZ;
        prevoff = currentoff, currentoff += entry->rec_len)
     {
-      entry = (struct ext2_dir_entry *)currentoff;
-      
+      entry = (struct ext2_dir_entry_2 *)currentoff;
+
       if (!entry->rec_len
 	  || entry->rec_len % EXT2_DIR_PAD
 	  || entry->name_len > EXT2_NAME_LEN
@@ -377,16 +413,16 @@ dirscanblock (vm_address_t blockaddr, struct node *dp, int idx, char *name,
 	  || EXT2_DIR_REC_LEN (entry->name_len) > entry->rec_len
 	  || memchr (entry->name, '\0', entry->name_len))
 	{
-	  ext2_warning ("bad directory entry: inode: %d offset: %ld",
+	  ext2_warning ("bad directory entry: inode: %Ld offset: %zd",
 			dp->cache_id,
 			currentoff - blockaddr + idx * DIRBLKSIZ);
 	  return ENOENT;
 	}
-      
+
       if (looking || countcopies)
 	{
-	  int thisfree;
-	  
+	  size_t thisfree;
+
 	  /* Count how much free space this entry has in it. */
 	  if (entry->inode == 0)
 	    thisfree = entry->rec_len;
@@ -398,9 +434,9 @@ dirscanblock (vm_address_t blockaddr, struct node *dp, int idx, char *name,
 	     number of bytes there too. */
 	  if (countcopies && currentoff != blockaddr)
 	    nbytes += EXT2_DIR_REC_LEN (entry->name_len);
-	  
+
 	  if (ds->stat == COMPRESS && nbytes > ds->nbytes)
-	    /* The previously found compress is better than 
+	    /* The previously found compress is better than
 	       this one, so don't bother counting any more. */
 	    countcopies = 0;
 
@@ -417,9 +453,9 @@ dirscanblock (vm_address_t blockaddr, struct node *dp, int idx, char *name,
 	      nfree += thisfree;
 	      if (nfree >= needed)
 		consider_compress = 1;
-	    }	      
+	    }
 	}
-      
+
       if (entry->inode)
 	nentries++;
 
@@ -430,17 +466,17 @@ dirscanblock (vm_address_t blockaddr, struct node *dp, int idx, char *name,
 	break;
     }
 
-  if (consider_compress 
-      && (ds->type == LOOKING
-	  || (ds->type == COMPRESS && ds->nbytes > nbytes)))
+  if (consider_compress
+      && (ds->stat == LOOKING
+	  || (ds->stat == COMPRESS && ds->nbytes > nbytes)))
     {
       ds->type = CREATE;
       ds->stat = COMPRESS;
-      ds->entry = (struct ext2_dir_entry *) blockaddr;
+      ds->entry = (struct ext2_dir_entry_2 *) blockaddr;
       ds->idx = idx;
       ds->nbytes = nbytes;
     }
-  
+
   if (currentoff >= blockaddr + DIRBLKSIZ)
     {
       int i;
@@ -450,9 +486,9 @@ dirscanblock (vm_address_t blockaddr, struct node *dp, int idx, char *name,
 	 down how many entries there were. */
       if (!dp->dn->dirents)
 	{
-	  dp->dn->dirents = malloc ((dp->dn_stat.st_size / DIRBLKSIZ) 
+	  dp->dn->dirents = malloc ((dp->dn_stat.st_size / DIRBLKSIZ)
 				    * sizeof (int));
-	  for (i = 0; i < dp->dn_stat.st_size/DIRBLKSIZ; i++) 
+	  for (i = 0; i < dp->dn_stat.st_size/DIRBLKSIZ; i++)
 	    dp->dn->dirents[i] = -1;
 	}
       /* Make sure the count is correct if there is one now. */
@@ -462,7 +498,7 @@ dirscanblock (vm_address_t blockaddr, struct node *dp, int idx, char *name,
 
       return ENOENT;
     }
-  
+
   /* We have found the required name. */
 
   if (ds && type == CREATE)
@@ -473,7 +509,7 @@ dirscanblock (vm_address_t blockaddr, struct node *dp, int idx, char *name,
       ds->stat = HERE_TIS;
       ds->entry = entry;
       ds->idx = idx;
-      ds->preventry = (struct ext2_dir_entry *) prevoff;
+      ds->preventry = (struct ext2_dir_entry_2 *) prevoff;
     }
 
   *inum = entry->inode;
@@ -487,53 +523,49 @@ dirscanblock (vm_address_t blockaddr, struct node *dp, int idx, char *name,
    only be made if the directory has been held locked continuously since
    the preceding lookup call, and only if that call returned ENOENT. */
 error_t
-diskfs_direnter_hard (struct node *dp, char *name, struct node *np,
+diskfs_direnter_hard (struct node *dp, const char *name, struct node *np,
 		      struct dirstat *ds, struct protid *cred)
 {
-  struct ext2_dir_entry *new;
-  int namelen = strlen (name);
-  int needed = EXT2_DIR_REC_LEN (namelen);
-  int oldneeded;
+  struct ext2_dir_entry_2 *new;
+  size_t namelen = strlen (name);
+  size_t needed = EXT2_DIR_REC_LEN (namelen);
+  size_t oldneeded;
   vm_address_t fromoff, tooff;
-  int totfreed;  
+  size_t totfreed;
   error_t err;
-  off_t oldsize = 0;
+  size_t oldsize = 0;
 
   assert (ds->type == CREATE);
-  
+
   assert (!diskfs_readonly);
 
   dp->dn_set_mtime = 1;
 
+  /* Select a location for the new directory entry.  Each branch of this
+     switch is responsible for setting NEW to point to the on-disk
+     directory entry being written, and setting NEW->rec_len appropriately.  */
+
   switch (ds->stat)
     {
     case TAKE:
       /* We are supposed to consume this slot. */
       assert (ds->entry->inode == 0 && ds->entry->rec_len >= needed);
-      
-      ds->entry->inode = np->cache_id;
-      ds->entry->name_len = namelen;
-      bcopy (name, ds->entry->name, namelen);
 
+      new = ds->entry;
       break;
-      
+
     case SHRINK:
       /* We are supposed to take the extra space at the end
 	 of this slot. */
       oldneeded = EXT2_DIR_REC_LEN (ds->entry->name_len);
       assert (ds->entry->rec_len - oldneeded >= needed);
-      
-      new = (struct ext2_dir_entry *) ((vm_address_t) ds->entry + oldneeded);
 
-      new->inode = np->cache_id;
+      new = (struct ext2_dir_entry_2 *) ((vm_address_t) ds->entry + oldneeded);
+
       new->rec_len = ds->entry->rec_len - oldneeded;
-      new->name_len = namelen;
-      bcopy (name, new->name, namelen);
-      
       ds->entry->rec_len = oldneeded;
-      
       break;
-      
+
     case COMPRESS:
       /* We are supposed to move all the entries to the
 	 front of the block, giving each the minimum
@@ -543,15 +575,15 @@ diskfs_direnter_hard (struct node *dp, char *name, struct node *np,
 
       while (fromoff < (vm_address_t) ds->entry + DIRBLKSIZ)
 	{
-	  struct ext2_dir_entry *from = (struct ext2_dir_entry *)fromoff;
-	  struct ext2_dir_entry *to = (struct ext2_dir_entry *) tooff;
-	  int fromreclen = from->rec_len;
+	  struct ext2_dir_entry_2 *from = (struct ext2_dir_entry_2 *)fromoff;
+	  struct ext2_dir_entry_2 *to = (struct ext2_dir_entry_2 *) tooff;
+	  size_t fromreclen = from->rec_len;
 
 	  if (from->inode != 0)
 	    {
 	      assert (fromoff >= tooff);
 
-	      bcopy (from, to, fromreclen);
+	      memmove (to, from, fromreclen);
 	      to->rec_len = EXT2_DIR_REC_LEN (to->name_len);
 
 	      tooff += to->rec_len;
@@ -561,51 +593,74 @@ diskfs_direnter_hard (struct node *dp, char *name, struct node *np,
 
       totfreed = (vm_address_t) ds->entry + DIRBLKSIZ - tooff;
       assert (totfreed >= needed);
-      
-      new = (struct ext2_dir_entry *) tooff;
-      new->inode = np->cache_id;
+
+      new = (struct ext2_dir_entry_2 *) tooff;
       new->rec_len = totfreed;
-      new->name_len = namelen;
-      bcopy (name, new->name, namelen);
       break;
 
     case EXTEND:
       /* Extend the file. */
       assert (needed <= DIRBLKSIZ);
-      
+
       oldsize = dp->dn_stat.st_size;
+      if ((off_t)(oldsize + DIRBLKSIZ) != (dp->dn_stat.st_size + DIRBLKSIZ))
+	{
+	  /* We can't possibly map the whole directory in.  */
+	  munmap ((caddr_t) ds->mapbuf, ds->mapextent);
+	  return EOVERFLOW;
+	}
       while (oldsize + DIRBLKSIZ > dp->allocsize)
 	{
 	  err = diskfs_grow (dp, oldsize + DIRBLKSIZ, cred);
 	  if (err)
 	    {
-	      vm_deallocate (mach_task_self (), ds->mapbuf, ds->mapextent);
+	      munmap ((caddr_t) ds->mapbuf, ds->mapextent);
 	      return err;
 	    }
 	}
 
-      new = (struct ext2_dir_entry *) (ds->mapbuf + oldsize);
+      new = (struct ext2_dir_entry_2 *) (ds->mapbuf + oldsize);
 
       dp->dn_stat.st_size = oldsize + DIRBLKSIZ;
       dp->dn_set_ctime = 1;
 
-      new->inode = np->cache_id;
       new->rec_len = DIRBLKSIZ;
-      new->name_len = namelen;
-      bcopy (name, new->name, namelen);
       break;
-      
+
     default:
-      assert (0);
+      new = 0;
+      assert (! "impossible: bogus status field in dirstat");
     }
 
+  /* NEW points to the directory entry being written, and its
+     rec_len field is already filled in.  Now fill in the rest.  */
+
+  new->inode = np->cache_id;
+#if 0
+  /* XXX We cannot enable this code because file types can change
+     (and conceivably quite often) with translator settings.
+     There is no way for the translator that determines the type of
+     the virtual node to cause all the directory entries linked to
+     its underlying inode to reflect the proper type.  */
+  new->file_type = (EXT2_HAS_INCOMPAT_FEATURE (sblock,
+					       EXT2_FEATURE_INCOMPAT_FILETYPE)
+		    ? file_type_ext2[IFTODT (np->dn_stat.st_mode & S_IFMT)]
+		    : 0);
+#else
+  new->file_type = 0;
+#endif
+  new->name_len = namelen;
+  memcpy (new->name, name, namelen);
+
+  /* Mark the directory inode has having been written.  */
+  dp->dn->info.i_flags &= ~EXT2_BTREE_FL;
   dp->dn_set_mtime = 1;
 
-  vm_deallocate (mach_task_self (), ds->mapbuf, ds->mapextent);
-        
+  munmap ((caddr_t) ds->mapbuf, ds->mapextent);
+
   if (ds->stat != EXTEND)
     {
-      /* If we are keeping count of this block, then keep the count up 
+      /* If we are keeping count of this block, then keep the count up
 	 to date. */
       if (dp->dn->dirents && dp->dn->dirents[ds->idx] != -1)
 	dp->dn->dirents[ds->idx]++;
@@ -617,10 +672,10 @@ diskfs_direnter_hard (struct node *dp, char *name, struct node *np,
 	 anything at all. */
       if (dp->dn->dirents)
 	{
-	  dp->dn->dirents = realloc (dp->dn->dirents, 
+	  dp->dn->dirents = realloc (dp->dn->dirents,
 				     (dp->dn_stat.st_size / DIRBLKSIZ
 				      * sizeof (int)));
-	  for (i = oldsize / DIRBLKSIZ; 
+	  for (i = oldsize / DIRBLKSIZ;
 	       i < dp->dn_stat.st_size / DIRBLKSIZ;
 	       i++)
 	    dp->dn->dirents[i] = -1;
@@ -629,15 +684,15 @@ diskfs_direnter_hard (struct node *dp, char *name, struct node *np,
 	}
       else
 	{
-	  dp->dn->dirents = malloc (dp->dn_stat.st_size / DIRBLKSIZ 
+	  dp->dn->dirents = malloc (dp->dn_stat.st_size / DIRBLKSIZ
 				    * sizeof (int));
 	  for (i = 0; i < dp->dn_stat.st_size / DIRBLKSIZ; i++)
 	    dp->dn->dirents[i] = -1;
 	  dp->dn->dirents[ds->idx] = 1;
 	}
     }
-  
-  diskfs_file_update (dp, 1);
+
+  diskfs_file_update (dp, diskfs_synchronous);
 
   return 0;
 }
@@ -645,17 +700,15 @@ diskfs_direnter_hard (struct node *dp, char *name, struct node *np,
 /* Following a lookup call for REMOVE, this removes the link from the
    directory.  DP is the directory being changed and DS is the cached
    information returned from lookup.  This call is only valid if the
-   directory has been locked continously since the call to lookup, and
+   directory has been locked continuously since the call to lookup, and
    only if that call succeeded.  */
 error_t
 diskfs_dirremove_hard (struct node *dp, struct dirstat *ds)
 {
   assert (ds->type == REMOVE);
   assert (ds->stat == HERE_TIS);
-  
-  assert (!diskfs_readonly);
 
-  dp->dn_set_mtime = 1;
+  assert (!diskfs_readonly);
 
   if (ds->preventry == 0)
     ds->entry->inode = 0;
@@ -667,23 +720,24 @@ diskfs_dirremove_hard (struct node *dp, struct dirstat *ds)
     }
 
   dp->dn_set_mtime = 1;
+  dp->dn->info.i_flags &= ~EXT2_BTREE_FL;
 
-  vm_deallocate (mach_task_self (), ds->mapbuf, ds->mapextent);
+  munmap ((caddr_t) ds->mapbuf, ds->mapextent);
 
   /* If we are keeping count of this block, then keep the count up
      to date. */
   if (dp->dn->dirents && dp->dn->dirents[ds->idx] != -1)
     dp->dn->dirents[ds->idx]--;
-  
-  diskfs_file_update (dp, 1);
+
+  diskfs_file_update (dp, diskfs_synchronous);
 
   return 0;
 }
-  
+
 
 /* Following a lookup call for RENAME, this changes the inode number
-   on a directory entry.  DP is the directory being changed; NP is 
-   the new node being linked in; DP is the cached information returned 
+   on a directory entry.  DP is the directory being changed; NP is
+   the new node being linked in; DP is the cached information returned
    by lookup.  This call is only valid if the directory has been locked
    continuously since the call to lookup, and only if that call
    succeeded.  */
@@ -692,15 +746,16 @@ diskfs_dirrewrite_hard (struct node *dp, struct node *np, struct dirstat *ds)
 {
   assert (ds->type == RENAME);
   assert (ds->stat == HERE_TIS);
-  
+
   assert (!diskfs_readonly);
 
   ds->entry->inode = np->cache_id;
   dp->dn_set_mtime = 1;
+  dp->dn->info.i_flags &= ~EXT2_BTREE_FL;
 
-  vm_deallocate (mach_task_self (), ds->mapbuf, ds->mapextent);
-  
-  diskfs_file_update (dp, 1);
+  munmap ((caddr_t) ds->mapbuf, ds->mapextent);
+
+  diskfs_file_update (dp, diskfs_synchronous);
 
   return 0;
 }
@@ -712,23 +767,26 @@ diskfs_dirempty (struct node *dp, struct protid *cred)
 {
   error_t err;
   vm_address_t buf = 0, curoff;
-  struct ext2_dir_entry *entry;
+  struct ext2_dir_entry_2 *entry;
   int hit = 0;			/* Found something in the directory.  */
   memory_object_t memobj = diskfs_get_filemap (dp, VM_PROT_READ);
-  
+
+  if (memobj == MACH_PORT_NULL)
+    /* XXX should reflect error properly. */
+    return 0;
+
   err = vm_map (mach_task_self (), &buf, dp->dn_stat.st_size, 0,
 		1, memobj, 0, 0, VM_PROT_READ, VM_PROT_READ, 0);
   mach_port_deallocate (mach_task_self (), memobj);
   assert (!err);
 
-  if (! diskfs_check_readonly ())
-    dp->dn_set_atime = 1;
+  diskfs_set_node_atime (dp);
 
-  for (curoff = buf; 
+  for (curoff = buf;
        !hit && curoff < buf + dp->dn_stat.st_size;
        curoff += entry->rec_len)
     {
-      entry = (struct ext2_dir_entry *) curoff;
+      entry = (struct ext2_dir_entry_2 *) curoff;
 
       if (entry->inode != 0
 	  && (entry->name_len > 2
@@ -738,12 +796,11 @@ diskfs_dirempty (struct node *dp, struct protid *cred)
 	hit = 1;
     }
 
-  if (! diskfs_check_readonly ())
-    dp->dn_set_atime = 1;
+  diskfs_set_node_atime (dp);
   if (diskfs_synchronous)
     diskfs_node_update (dp, 1);
 
-  vm_deallocate (mach_task_self (), buf, dp->dn_stat.st_size);
+  munmap ((caddr_t) buf, dp->dn_stat.st_size);
 
   return !hit;
 }
@@ -755,7 +812,7 @@ diskfs_drop_dirstat (struct node *dp, struct dirstat *ds)
   if (ds->type != LOOKUP)
     {
       assert (ds->mapbuf);
-      vm_deallocate (mach_task_self (), ds->mapbuf, ds->mapextent);
+      munmap ((caddr_t) ds->mapbuf, ds->mapextent);
       ds->type = LOOKUP;
     }
   return 0;
@@ -766,31 +823,31 @@ diskfs_drop_dirstat (struct node *dp, struct dirstat *ds)
    write the answer down in its dirents array.  As a side affect
    fill BUF with the block.  */
 static error_t
-count_dirents (struct node *dp, int nb, char *buf)
+count_dirents (struct node *dp, block_t nb, char *buf)
 {
-  int amt;
+  size_t amt;
   char *offinblk;
-  struct ext2_dir_entry *entry;
+  struct ext2_dir_entry_2 *entry;
   int count = 0;
   error_t err;
 
   assert (dp->dn->dirents);
   assert ((nb + 1) * DIRBLKSIZ <= dp->dn_stat.st_size);
-  
+
   err = diskfs_node_rdwr (dp, buf, nb * DIRBLKSIZ, DIRBLKSIZ, 0, 0, &amt);
   if (err)
     return err;
   assert (amt == DIRBLKSIZ);
-  
+
   for (offinblk = buf;
        offinblk < buf + DIRBLKSIZ;
        offinblk += entry->rec_len)
     {
-      entry = (struct ext2_dir_entry *) offinblk;
+      entry = (struct ext2_dir_entry_2 *) offinblk;
       if (entry->inode)
 	count++;
     }
-  
+
   assert (dp->dn->dirents[nb] == -1 || dp->dn->dirents[nb] == count);
   dp->dn->dirents[nb] = count;
   return 0;
@@ -800,19 +857,19 @@ count_dirents (struct node *dp, int nb, char *buf)
    Must be a power of two.  */
 #define DIRENT_ALIGN 4
 
-/* Implement the disikfs_get_directs callback as described in
+/* Implement the diskfs_get_directs callback as described in
    <hurd/diskfs.h>. */
 error_t
-diskfs_get_directs (struct node *dp, 
-		    int entry, 
+diskfs_get_directs (struct node *dp,
+		    int entry,
 		    int nentries,
 		    char **data,
-		    u_int *datacnt,
+		    size_t *datacnt,
 		    vm_size_t bufsiz,
 		    int *amt)
 {
-  int blkno;
-  int nblks;
+  block_t blkno;
+  block_t nblks;
   int curentry;
   char buf[DIRBLKSIZ];
   char *bufp;
@@ -820,11 +877,11 @@ diskfs_get_directs (struct node *dp,
   error_t err;
   int i;
   char *datap;
-  struct ext2_dir_entry *entryp;
+  struct ext2_dir_entry_2 *entryp;
   int allocsize;
-  int checklen;
+  size_t checklen;
   struct dirent *userp;
-  
+
   nblks = dp->dn_stat.st_size/DIRBLKSIZ;
 
   if (!dp->dn->dirents)
@@ -834,6 +891,40 @@ diskfs_get_directs (struct node *dp,
 	dp->dn->dirents[i] = -1;
     }
 
+  /* Scan through the entries to find ENTRY.  If we encounter
+     a -1 in the process then stop to fill it.  When we run
+     off the end, ENTRY is too big. */
+  curentry = 0;
+  bufvalid = 0;
+  for (blkno = 0; blkno < nblks; blkno++)
+    {
+      if (dp->dn->dirents[blkno] == -1)
+	{
+	  err = count_dirents (dp, blkno, buf);
+	  if (err)
+	    return err;
+	  bufvalid = 1;
+	}
+
+      if (curentry + dp->dn->dirents[blkno] > entry)
+	/* ENTRY starts in this block. */
+	break;
+
+      curentry += dp->dn->dirents[blkno];
+
+      bufvalid = 0;
+    }
+
+  if (blkno == nblks)
+    {
+      /* We reached the end of the directory without seeing ENTRY.
+	 This is treated as an EOF condition, meaning we return
+	 success with empty results.  */
+      *datacnt = 0;
+      *amt = 0;
+      return 0;
+    }
+
   /* Allocate enough space to hold the maximum we might return */
   if (!bufsiz || bufsiz > dp->dn_stat.st_size)
     /* Allocate enough to return the entire directory.  Since ext2's
@@ -859,38 +950,7 @@ diskfs_get_directs (struct node *dp,
     allocsize = round_page (bufsiz);
 
   if (allocsize > *datacnt)
-    vm_allocate (mach_task_self (), (vm_address_t *) data, allocsize, 1);
-  
-  /* Scan through the entries to find ENTRY.  If we encounter
-     a -1 in the process then stop to fill it.  When we run
-     off the end, ENTRY is too big. */
-  curentry = 0;
-  bufvalid = 0;
-  for (blkno = 0; blkno < nblks; blkno++)
-    {
-      if (dp->dn->dirents[blkno] == -1)
-	{
-	  err = count_dirents (dp, blkno, buf);
-	  if (err)
-	    return err;
-	  bufvalid = 1;
-	}
-
-      if (curentry + dp->dn->dirents[blkno] > entry)
-	/* ENTRY starts in this block. */
-	break;
-
-      curentry += dp->dn->dirents[blkno];
-      
-      bufvalid = 0;
-    }
-  
-  if (blkno == nblks)
-    {
-      *datacnt = 0;
-      *amt = 0;
-      return 0;
-    }
+    *data = mmap (0, allocsize, PROT_READ|PROT_WRITE, MAP_ANON, 0, 0);
 
   /* Set bufp appropriately */
   bufp = buf;
@@ -907,13 +967,13 @@ diskfs_get_directs (struct node *dp,
 	  assert (checklen == DIRBLKSIZ);
 	  bufvalid = 1;
 	}
-      for (i = 0, bufp = buf; 
-	   i < entry - curentry && bufp - buf < DIRBLKSIZ; 
-	   bufp += ((struct ext2_dir_entry *)bufp)->rec_len, i++)
+      for (i = 0, bufp = buf;
+	   i < entry - curentry && bufp - buf < DIRBLKSIZ;
+	   bufp += ((struct ext2_dir_entry_2 *)bufp)->rec_len, i++)
 	;
       /* Make sure we didn't run off the end. */
       assert (bufp - buf < DIRBLKSIZ);
-    }      
+    }
 
   i = 0;
   datap = *data;
@@ -925,7 +985,7 @@ diskfs_get_directs (struct node *dp,
     {
       if (!bufvalid)
 	{
-	  err = diskfs_node_rdwr (dp, buf, blkno * DIRBLKSIZ, DIRBLKSIZ, 
+	  err = diskfs_node_rdwr (dp, buf, blkno * DIRBLKSIZ, DIRBLKSIZ,
 				  0, 0, &checklen);
 	  if (err)
 	    return err;
@@ -934,7 +994,7 @@ diskfs_get_directs (struct node *dp,
 	  bufp = buf;
 	}
 
-      entryp = (struct ext2_dir_entry *)bufp;
+      entryp = (struct ext2_dir_entry_2 *)bufp;
 
       if (entryp->inode)
 	{
@@ -963,7 +1023,30 @@ diskfs_get_directs (struct node *dp,
 	  userp->d_fileno = entryp->inode;
 	  userp->d_reclen = rec_len;
 	  userp->d_namlen = name_len;
-	  bcopy (entryp->name, userp->d_name, name_len);
+
+#if 0
+	  /* We don't bother to check the EXT2_FEATURE_INCOMPAT_FILETYPE
+	     flag in the superblock, because in old filesystems the
+	     file_type field is the high byte of the length field and is
+	     always zero because names cannot be that long.  */
+	  if (entryp->file_type < EXT2_FT_MAX)
+	    userp->d_type = ext2_file_type[entryp->file_type];
+	  else
+	    {
+	      ext2_warning ("bad type %d in directory entry: "
+			    "inode: %d offset: %d",
+			    entryp->file_type,
+			    dp->cache_id,
+			    blkno * DIRBLKSIZ + bufp - buf);
+	      userp->d_type = DT_UNKNOWN;
+	    }
+#else
+	  /* XXX
+	     For complex reasons it might not be correct to return
+	     the filesystem's d_type value to the user.  */
+	  userp->d_type = DT_UNKNOWN;
+#endif
+	  memcpy (userp->d_name, entryp->name, name_len);
 	  userp->d_name[name_len] = '\0';
 
 	  datap += rec_len;
@@ -972,7 +1055,7 @@ diskfs_get_directs (struct node *dp,
 
       if (entryp->rec_len == 0)
 	{
-	  ext2_warning ("zero length directory entry: inode: %d offset: %d",
+	  ext2_warning ("zero length directory entry: inode: %Ld offset: %zd",
 			dp->cache_id,
 			blkno * DIRBLKSIZ + bufp - buf);
 	  return EIO;
@@ -986,23 +1069,22 @@ diskfs_get_directs (struct node *dp,
 	}
       else if (bufp - buf > DIRBLKSIZ)
 	{
-	  ext2_warning ("directory entry too long: inode: %d offset: %d",
+	  ext2_warning ("directory entry too long: inode: %Ld offset: %zd",
 			dp->cache_id,
 			blkno * DIRBLKSIZ + bufp - buf - entryp->rec_len);
 	  return EIO;
 	}
     }
-  
+
   /* We've copied all we can.  If we allocated our own array
      but didn't fill all of it, then free whatever memory we didn't use. */
   if (allocsize > *datacnt)
     {
       if (round_page (datap - *data) < allocsize)
-	vm_deallocate (mach_task_self (), 
-		       (vm_address_t) (*data + round_page (datap - *data)),
-		       allocsize - round_page (datap - *data));
+	munmap ((caddr_t) (*data + round_page (datap - *data)),
+		allocsize - round_page (datap - *data));
     }
-  
+
   /* Set variables for return */
   *datacnt = datap - *data;
   *amt = i;
