@@ -1,8 +1,8 @@
 /* File truncation
 
-   Copyright (C) 1995, 1996 Free Software Foundation, Inc.
+   Copyright (C) 1995,96,97,99,2000 Free Software Foundation, Inc.
 
-   Written by Miles Bader <miles@gnu.ai.mit.edu>
+   Written by Miles Bader <miles@gnu.org>
 
    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
@@ -102,7 +102,7 @@ trunc_direct (struct node *node, block_t end, struct free_block_run *fbr)
 {
   block_t *blocks = node->dn->info.i_data;
 
-  ext2_debug ("truncating direct blocks from %ld", end);
+  ext2_debug ("truncating direct blocks from %d", end);
 
   while (end < EXT2_NDIR_BLOCKS)
     free_block_run_free_ptr (fbr, blocks + end++);
@@ -124,7 +124,7 @@ trunc_indirect (struct node *node, block_t end,
     {
       unsigned index;
       int modified = 0, all_freed = 1;
-      block_t *ind_bh = (block_t *)bptr (*p);
+      block_t *ind_bh = (block_t *) disk_cache_block_ref (*p);
       unsigned first = end < offset ? 0 : end - offset;
 
       for (index = first; index < addr_per_block; index++)
@@ -138,9 +138,17 @@ trunc_indirect (struct node *node, block_t end,
 	  }
 
       if (first == 0 && all_freed)
-	free_block_run_free_ptr (fbr, p);
+	{
+	  pager_flush_some (diskfs_disk_pager,
+			    bptr_index (ind_bh) << log2_block_size,
+			    block_size, 1);
+	  free_block_run_free_ptr (fbr, p);
+	  disk_cache_block_deref (ind_bh);
+	}
       else if (modified)
 	record_indir_poke (node, ind_bh);
+      else
+	disk_cache_block_deref (ind_bh);
     }
 }
 
@@ -205,7 +213,7 @@ poke_pages (memory_object_t obj, vm_offset_t start, vm_offset_t end)
 	  vm_address_t poke;
 	  for (poke = addr; poke < addr + len; poke += vm_page_size)
 	    *(volatile int *)poke = *(volatile int *)poke;
-	  vm_deallocate (mach_task_self (), addr, len);
+	  munmap ((caddr_t) addr, len);
 	}
 
       start += len;
@@ -221,21 +229,27 @@ force_delayed_copies (struct node *node, off_t length)
 {
   struct pager *pager;
 
-  spin_lock (&node_to_page_lock);
+  pthread_spin_lock (&node_to_page_lock);
   pager = node->dn->pager;
   if (pager)
     ports_port_ref (pager);
-  spin_unlock (&node_to_page_lock);
- 
+  pthread_spin_unlock (&node_to_page_lock);
+
   if (pager)
     {
       mach_port_t obj;
-      
+
       pager_change_attributes (pager, MAY_CACHE, MEMORY_OBJECT_COPY_NONE, 1);
       obj = diskfs_get_filemap (node, VM_PROT_READ);
-      poke_pages (obj, round_page (length), round_page (node->allocsize));
-      mach_port_deallocate (mach_task_self (), obj);
-      pager_flush_some (pager, round_page(length), node->allocsize - length, 1);
+      if (obj != MACH_PORT_NULL)
+	{
+	  /* XXX should cope with errors from diskfs_get_filemap */
+	  poke_pages (obj, round_page (length), round_page (node->allocsize));
+	  mach_port_deallocate (mach_task_self (), obj);
+	  pager_flush_some (pager, round_page(length),
+			    node->allocsize - length, 1);
+	}
+
       ports_port_deref (pager);
     }
 }
@@ -245,11 +259,11 @@ enable_delayed_copies (struct node *node)
 {
   struct pager *pager;
 
-  spin_lock (&node_to_page_lock);
+  pthread_spin_lock (&node_to_page_lock);
   pager = node->dn->pager;
   if (pager)
     ports_port_ref (pager);
-  spin_unlock (&node_to_page_lock);
+  pthread_spin_unlock (&node_to_page_lock);
 
   if (pager)
     {
@@ -263,7 +277,7 @@ enable_delayed_copies (struct node *node)
 /* The user must define this function.  Truncate locked node NODE to be SIZE
    bytes long.  (If NODE is already less than or equal to SIZE bytes
    long, do nothing.)  If this is a symlink (and diskfs_shortcut_symlink
-   is set) then this should clear the symlink, even if 
+   is set) then this should clear the symlink, even if
    diskfs_create_symlink_hook stores the link target elsewhere.  */
 error_t
 diskfs_truncate (struct node *node, off_t length)
@@ -285,7 +299,7 @@ diskfs_truncate (struct node *node, off_t length)
       node->dn_stat.st_size = length;
       node->dn_set_mtime = 1;
       node->dn_set_ctime = 1;
-      diskfs_node_update (node, 1);
+      diskfs_node_update (node, diskfs_synchronous);
       return 0;
     }
 
@@ -295,28 +309,29 @@ diskfs_truncate (struct node *node, off_t length)
    * zeroed in case it ever becomes accessible again because of
    * subsequent file growth.
    */
-  offset = length % block_size;
+  offset = length & (block_size - 1);
   if (offset > 0)
     {
       diskfs_node_rdwr (node, (void *)zeroblock, length, block_size - offset,
 			1, 0, 0);
+      /* Make sure that really happens to avoid leaks.  */
       diskfs_file_update (node, 1);
     }
 
-  ext2_discard_prealloc(node);
+  ext2_discard_prealloc (node);
 
   force_delayed_copies (node, length);
 
-  rwlock_writer_lock (&node->dn->alloc_lock);
+  pthread_rwlock_wrlock (&node->dn->alloc_lock);
 
   /* Update the size on disk; fsck will finish freeing blocks if necessary
      should we crash. */
   node->dn_stat.st_size = length;
   node->dn_set_mtime = 1;
   node->dn_set_ctime = 1;
-  diskfs_node_update (node, 1);
+  diskfs_node_update (node, diskfs_synchronous);
 
-  err = diskfs_catch_exception();
+  err = diskfs_catch_exception ();
   if (!err)
     {
       block_t end = boffs_block (round_block (length)), offs;
@@ -342,6 +357,8 @@ diskfs_truncate (struct node *node, off_t length)
 	 won't hurt if is wrong.  */
       node->dn->last_page_partially_writable =
 	trunc_page (node->allocsize) != node->allocsize;
+
+      diskfs_end_catch_exception ();
     }
 
   node->dn_set_mtime = 1;
@@ -351,7 +368,7 @@ diskfs_truncate (struct node *node, off_t length)
   /* Now we can permit delayed copies again. */
   enable_delayed_copies (node);
 
-  rwlock_writer_unlock (&node->dn->alloc_lock);
+  pthread_rwlock_unlock (&node->dn->alloc_lock);
 
   return err;
 }
