@@ -1,5 +1,5 @@
-/* 
-   Copyright (C) 1996, 1997 Free Software Foundation, Inc.
+/*
+   Copyright (C) 1996,97,2001,02 Free Software Foundation, Inc.
    Written by Michael I. Bushnell, p/BSG.
 
    This file is part of the GNU Hurd.
@@ -25,7 +25,9 @@
 #include <fcntl.h>
 
 error_t
-netfs_S_fsys_getroot (mach_port_t cntl,
+netfs_S_fsys_getroot (struct netfs_control *pt,
+		      mach_port_t reply,
+		      mach_msg_type_name_t reply_type,
 		      mach_port_t dotdot,
 		      uid_t *uids, mach_msg_type_number_t nuids,
 		      uid_t *gids, mach_msg_type_number_t ngids,
@@ -33,38 +35,35 @@ netfs_S_fsys_getroot (mach_port_t cntl,
 		      retry_type *do_retry,
 		      char *retry_name,
 		      mach_port_t *retry_port,
-		      mach_port_t *retry_port_type)
+		      mach_msg_type_name_t *retry_port_type)
 {
-  struct port_info *pt = ports_lookup_port (netfs_port_bucket, cntl,
-					    netfs_control_class);
   struct iouser *cred;
   error_t err;
   struct protid *newpi;
   mode_t type;
-  struct idvec *uvec, *gvec;
-  struct peropen peropen_context = { root_parent: dotdot };
+  struct peropen peropen_context =
+    {
+      root_parent: dotdot,
+      path: NULL,
+    };
 
   if (!pt)
     return EOPNOTSUPP;
-  ports_port_deref (pt);
 
-  uvec = make_idvec ();
-  gvec = make_idvec ();
-  idvec_set_ids (uvec, uids, nuids);
-  idvec_set_ids (gvec, gids, ngids);
+  err = iohelp_create_complex_iouser (&cred, uids, nuids, gids, ngids);
+  if (err)
+    return err;
 
-  cred = iohelp_create_iouser (uvec, gvec);
-  
   flags &= O_HURD;
-  
-  mutex_lock (&netfs_root_node->lock);
+
+  pthread_mutex_lock (&netfs_root_node->lock);
   err = netfs_validate_stat (netfs_root_node, cred);
   if (err)
     goto out;
-  
+
   type = netfs_root_node->nn_stat.st_mode & S_IFMT;
 
-  if ((netfs_root_node->istranslated
+  if (((netfs_root_node->nn_stat.st_mode & S_IPTRANS)
        || fshelp_translated (&netfs_root_node->transbox))
       && !(flags & O_NOTRANS))
     {
@@ -75,7 +74,7 @@ netfs_S_fsys_getroot (mach_port_t cntl,
 			       do_retry, retry_name, retry_port);
       if (err != ENOENT)
 	{
-	  mutex_unlock (&netfs_root_node->lock);
+	  pthread_mutex_unlock (&netfs_root_node->lock);
 	  iohelp_free_iouser (cred);
 	  if (!err)
 	    *retry_port_type = MACH_MSG_TYPE_MOVE_SEND;
@@ -84,17 +83,17 @@ netfs_S_fsys_getroot (mach_port_t cntl,
       /* ENOENT means translator has vanished inside fshelp_fetch_root. */
       err = 0;
     }
-  
+
   if (type == S_IFLNK && !(flags & (O_NOLINK | O_NOTRANS)))
     {
       char pathbuf[netfs_root_node->nn_stat.st_size + 1];
-      
+
       err = netfs_attempt_readlink (cred, netfs_root_node, pathbuf);
 
       if (err)
 	goto out;
-      
-      mutex_unlock (&netfs_root_node->lock);
+
+      pthread_mutex_unlock (&netfs_root_node->lock);
       iohelp_free_iouser (cred);
 
       if (pathbuf[0] == '/')
@@ -115,20 +114,20 @@ netfs_S_fsys_getroot (mach_port_t cntl,
 	  return 0;
 	}
     }
-  
-  if ((type == S_IFSOCK || type == S_IFBLK || type == S_IFCHR 
+
+  if ((type == S_IFSOCK || type == S_IFBLK || type == S_IFCHR
       || type == S_IFIFO) && (flags & (O_READ|O_WRITE|O_EXEC)))
     {
       err = EOPNOTSUPP;
       goto out;
     }
-  
+
   err = netfs_check_open_permissions (cred, netfs_root_node, flags, 0);
   if (err)
     goto out;
-  
+
   flags &= ~OPENONLY_STATE_MODES;
-  
+
   newpi = netfs_make_protid (netfs_make_peropen (netfs_root_node, flags,
 						 &peropen_context),
 			     cred);
@@ -139,13 +138,10 @@ netfs_S_fsys_getroot (mach_port_t cntl,
   *retry_port_type = MACH_MSG_TYPE_MAKE_SEND;
   retry_name[0] = '\0';
   ports_port_deref (newpi);
-  
+
  out:
   if (err)
     iohelp_free_iouser (cred);
-  mutex_unlock (&netfs_root_node->lock);
+  pthread_mutex_unlock (&netfs_root_node->lock);
   return err;
 }
-
-  
-      
