@@ -1,5 +1,5 @@
-/* 
-   Copyright (C) 1995, 1996, 1997 Free Software Foundation, Inc.
+/*
+   Copyright (C) 1995, 1996, 1997, 1999 Free Software Foundation, Inc.
    Written by Michael I. Bushnell, p/BSG.
 
    This file is part of the GNU Hurd.
@@ -21,6 +21,7 @@
 #include "netfs.h"
 #include "io_S.h"
 #include <fcntl.h>
+#include <sys/mman.h>
 
 error_t
 netfs_S_io_read (struct protid *user,
@@ -38,18 +39,18 @@ netfs_S_io_read (struct protid *user,
     return EOPNOTSUPP;
 
   node = user->po->np;
-  mutex_lock (&user->po->np->lock);
+  pthread_mutex_lock (&user->po->np->lock);
 
   if ((user->po->openstat & O_READ) == 0)
     {
-      mutex_unlock (&node->lock);
+      pthread_mutex_unlock (&node->lock);
       return EBADF;
     }
 
   if (amount > *datalen)
     {
       alloced = 1;
-      vm_allocate (mach_task_self (), (vm_address_t *) data, amount, 1);
+      *data = mmap (0, amount, PROT_READ|PROT_WRITE, MAP_ANON, 0, 0);
     }
   *datalen = amount;
 
@@ -64,6 +65,8 @@ netfs_S_io_read (struct protid *user,
 
       if (start + amount > size)
 	amount = size - start;
+      if (amount > size)
+	amount = size;
 
       if (start >= size)
 	{
@@ -73,7 +76,7 @@ netfs_S_io_read (struct protid *user,
       else if (amount < size || start > 0)
 	{
 	  char *whole_link = alloca (size);
-	  err = netfs_attempt_readlink (user->user, node, *data);
+	  err = netfs_attempt_readlink (user->user, node, whole_link);
 	  if (! err)
 	    {
 	      memcpy (*data, whole_link + start, amount);
@@ -81,7 +84,10 @@ netfs_S_io_read (struct protid *user,
 	    }
 	}
       else
-	err = netfs_attempt_readlink (user->user, node, *data);
+	{
+	  err = netfs_attempt_readlink (user->user, node, *data);
+	  *datalen = amount;
+	}
     }
   else
     /* Read from a normal file.  */
@@ -90,16 +96,14 @@ netfs_S_io_read (struct protid *user,
   if (offset == -1 && !err)
     user->po->filepointer += *datalen;
 
-  mutex_unlock (&node->lock);
+  pthread_mutex_unlock (&node->lock);
 
   if (err && alloced)
-    vm_deallocate (mach_task_self (), (vm_address_t) *data, amount);
+    munmap (*data, amount);
 
   if (!err && alloced && (round_page (*datalen) < round_page (amount)))
-    vm_deallocate (mach_task_self (), 
-		   (vm_address_t) *data + round_page (*datalen),
-		   round_page (amount) - round_page (*datalen));
+    munmap (*data + round_page (*datalen),
+	    round_page (amount) - round_page (*datalen));
 
   return err;
 }
-
