@@ -1,5 +1,5 @@
 /* 
-   Copyright (C) 1995, 1996 Free Software Foundation, Inc.
+   Copyright (C) 1995, 1996, 2000 Free Software Foundation, Inc.
    Written by Michael I. Bushnell, p/BSG.
 
    This file is part of the GNU Hurd.
@@ -36,17 +36,14 @@ netfs_S_io_write (struct protid *user,
   if (!user)
     return EOPNOTSUPP;
   
-  np = user->po->np;
-
-  mutex_lock (&np->lock);
   if ((user->po->openstat & O_WRITE) == 0)
-    {
-      mutex_unlock (&np->lock);
-      return EBADF;
-    }
+    return EBADF;
 
   *amount = datalen;
 
+  np = user->po->np;
+  pthread_mutex_lock (&np->lock);
+
   if (off == -1)
     {
       if (user->po->openstat & O_APPEND)
@@ -54,7 +51,7 @@ netfs_S_io_write (struct protid *user,
 	  err = netfs_validate_stat (np, user->user);
 	  if (err)
 	    {
-	      mutex_unlock (&np->lock);
+	      pthread_mutex_unlock (&np->lock);
 	      return err;
 	    }
 	  user->po->filepointer = np->nn_stat.st_size;
@@ -65,9 +62,7 @@ netfs_S_io_write (struct protid *user,
   err =  netfs_attempt_write (user->user, np, off, amount, data);
   if (offset == -1 && !err)
     user->po->filepointer += *amount;
-  mutex_unlock (&np->lock);
+  pthread_mutex_unlock (&np->lock);
   
   return err;
 }
-
-
