@@ -19,7 +19,6 @@
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */
 
 #include "ports.h"
-#include <cthreads.h>
 #include <hurd/ihash.h>
 
 void *
@@ -27,29 +26,20 @@ ports_lookup_port (struct port_bucket *bucket,
 		   mach_port_t port,
 		   struct port_class *class)
 {
-  struct port_info *pi = 0;
-  
-  mutex_lock (&_ports_lock);
+  struct port_info *pi;
 
-  if (bucket)
-    pi = ihash_find (bucket->htable, port);
-  else
-    for (bucket = _ports_all_buckets; bucket; bucket = bucket->next)
-      {
-	pi = ihash_find (bucket->htable, port);
-	if (pi)
-	  break;
-      }
-  
-  if (pi && class && pi->class != class)
+  pthread_rwlock_rdlock (&_ports_htable_lock);
+
+  pi = hurd_ihash_find (&_ports_htable, port);
+  if (pi
+      && ((class && pi->class != class)
+          || (bucket && pi->bucket != bucket)))
     pi = 0;
 
   if (pi)
-    pi->refcnt++;
+    refcounts_unsafe_ref (&pi->refcounts, NULL);
+
+  pthread_rwlock_unlock (&_ports_htable_lock);
 
-  mutex_unlock (&_ports_lock);
-  
   return pi;
 }
-
-  
