@@ -1,8 +1,8 @@
 /* fsys_forward
 
-   Copyright (C) 1995 Free Software Foundation, Inc.
+   Copyright (C) 1995,2002 Free Software Foundation, Inc.
 
-   Written by Miles Bader <miles@gnu.ai.mit.edu>
+   Written by Miles Bader <miles@gnu.org>
 
    This file is part of the GNU Hurd.
 
@@ -11,7 +11,7 @@
    the Free Software Foundation; either version 2, or (at your option)
    any later version.
 
-   The GNU Hurd is distributed in the hope that it will be useful, 
+   The GNU Hurd is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
@@ -21,7 +21,7 @@
    the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */
 
 #include "priv.h"
-#include "fsys_S.h"
+#include "trivfs_fsys_S.h"
 
 /* Ask SERVER to provide fsys translation service for us.  REQUESTOR is
    the bootstrap port supplied to the original translator, and ARGV are
