@@ -1,31 +1,31 @@
 /*
-   Copyright (C) 1993, 1994, 1995, 1996 Free Software Foundation
+   Copyright (C) 1993,94,95,96,2001,02 Free Software Foundation, Inc.
 
-This file is part of the GNU Hurd.
+   This file is part of the GNU Hurd.
 
-The GNU Hurd is free software; you can redistribute it and/or modify
-it under the terms of the GNU General Public License as published by
-the Free Software Foundation; either version 2, or (at your option)
-any later version.
+   The GNU Hurd is free software; you can redistribute it and/or modify
+   it under the terms of the GNU General Public License as published by
+   the Free Software Foundation; either version 2, or (at your option)
+   any later version.
 
-The GNU Hurd is distributed in the hope that it will be useful, 
-but WITHOUT ANY WARRANTY; without even the implied warranty of
-MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-GNU General Public License for more details.
+   The GNU Hurd is distributed in the hope that it will be useful,
+   but WITHOUT ANY WARRANTY; without even the implied warranty of
+   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+   GNU General Public License for more details.
 
-You should have received a copy of the GNU General Public License
-along with the GNU Hurd; see the file COPYING.  If not, write to
-the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */
+   You should have received a copy of the GNU General Public License
+   along with the GNU Hurd; see the file COPYING.  If not, write to
+   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */
 
 /* Written by Michael I. Bushnell.  */
 
 #include "priv.h"
-#include "io_S.h"
+#include "trivfs_io_S.h"
 #include <string.h>
 
 /* Tell if the array LIST (of size N) contains a member equal to QUERY. */
 static inline int
-listmember (int *list, int query, int n)
+listmember (const uid_t *list, uid_t query, int n)
 {
   int i;
   for (i = 0; i < n; i++)
@@ -40,43 +40,68 @@ trivfs_S_io_restrict_auth (struct trivfs_protid *cred,
 			   mach_msg_type_name_t replytype,
 			   mach_port_t *newport,
 			   mach_msg_type_name_t *newporttype,
-			   uid_t *uids, u_int nuids,
-			   uid_t *gids, u_int ngids)
+			   uid_t *uids, size_t nuids,
+			   uid_t *gids, size_t ngids)
 {
-  int i;
-  error_t err = 0;
+  unsigned int i;
+  error_t err;
   struct trivfs_protid *newcred;
   struct idvec *uvec, *gvec;
   struct iouser *user;
-  
+
   if (!cred)
     return EOPNOTSUPP;
-  
-  uvec = make_idvec ();
-  gvec = make_idvec ();
-
-  user = iohelp_create_iouser (uvec, gvec);
 
   if (cred->isroot)
+    /* CRED has root access, and so may use any ids.  */
     {
-      /* CRED has root access, and so may use any ids.  */
-      idvec_set_ids (uvec, uids, nuids);
-      idvec_set_ids (gvec, gids, ngids);
+      err = iohelp_create_complex_iouser (&user, uids, nuids, gids, ngids);
+      if (err)
+        return err;
     }
   else
     {
+      uvec = make_idvec ();
+      if (! uvec)
+        return ENOMEM;
+
+      gvec = make_idvec ();
+      if (! gvec)
+        {
+	  idvec_free (uvec);
+	  return ENOMEM;
+	}
+
       /* Otherwise, use any of the requested ids that CRED already has.  */
       for (i = 0; i < cred->user->uids->num; i++)
 	if (listmember (uids, cred->user->uids->ids[i], nuids))
-	  idvec_add (uvec, cred->user->uids->ids[i]);
+	  {
+	    err = idvec_add (uvec, cred->user->uids->ids[i]);
+	    if (err)
+	      goto out;
+	  }
+
       for (i = 0; i < cred->user->gids->num; i++)
 	if (listmember (gids, cred->user->gids->ids[i], ngids))
-	  idvec_add (gvec, cred->user->gids->ids[i]);
+	  {
+	    err = idvec_add (gvec, cred->user->gids->ids[i]);
+	    if (err)
+	      goto out;
+	  }
+
+      err = iohelp_create_iouser (&user, uvec, gvec);
+      if (err)
+        {
+	out:
+	  idvec_free (uvec);
+	  idvec_free (gvec);
+	  return err;
+	}
     }
 
   err = ports_create_port (cred->po->cntl->protid_class,
 			   cred->po->cntl->protid_bucket,
-			   sizeof (struct trivfs_protid), 
+			   sizeof (struct trivfs_protid),
 			   &newcred);
   if (err)
     {
@@ -85,17 +110,15 @@ trivfs_S_io_restrict_auth (struct trivfs_protid *cred,
     }
 
   newcred->isroot = 0;
-  mutex_lock (&cred->po->cntl->lock);
   newcred->po = cred->po;
-  newcred->po->refcnt++;
-  mutex_unlock (&cred->po->cntl->lock);
-  if (cred->isroot && idvec_contains (uvec, 0))
+  refcount_ref (&newcred->po->refcnt);
+  if (cred->isroot && idvec_contains (user->uids, 0))
     newcred->isroot = 1;
   newcred->user = user;
   newcred->hook = cred->hook;
 
-  err = io_restrict_auth (cred->realnode, &newcred->realnode, 
-			  user->uids->ids, user->uids->num, 
+  err = io_restrict_auth (cred->realnode, &newcred->realnode,
+			  user->uids->ids, user->uids->num,
 			  user->gids->ids, user->gids->num);
   if (!err && trivfs_protid_create_hook)
     {
