@@ -21,13 +21,16 @@
 #define IO_SELECT_REPLY_PORT
 
 #define IO_INTRAN sock_user_t begin_using_sock_user_port (io_t)
+#define IO_INTRAN_PAYLOAD sock_user_t begin_using_sock_user_payload
 #define IO_DESTRUCTOR end_using_sock_user_port (sock_user_t)
 
 #define IO_IMPORTS import "mig-decls.h";
 
 #define SOCKET_INTRAN sock_user_t begin_using_sock_user_port (socket_t)
+#define SOCKET_INTRAN_PAYLOAD sock_user_t begin_using_sock_user_payload
 #define SOCKET_DESTRUCTOR end_using_sock_user_port (sock_user_t)
 #define ADDRPORT_INTRAN addr_t begin_using_addr_port (addr_port_t)
+#define ADDRPORT_INTRAN_PAYLOAD addr_t begin_using_addr_payload
 #define ADDRPORT_DESTRUCTOR end_using_addr_port (addr_t)
 
 #define SOCKET_IMPORTS import "mig-decls.h";
