@@ -3,21 +3,21 @@
    Copyright (C) 1992, 1993, 1994, 1996, 1997, 2001, 2002, 2013 Free Software
    Foundation, Inc.
 
-This file is part of the GNU Hurd.
+   This file is part of the GNU Hurd.
 
-The GNU Hurd is free software; you can redistribute it and/or modify
-it under the terms of the GNU General Public License as published by
-the Free Software Foundation; either version 2, or (at your option)
-any later version.
+   The GNU Hurd is free software; you can redistribute it and/or modify
+   it under the terms of the GNU General Public License as published by
+   the Free Software Foundation; either version 2, or (at your option)
+   any later version.
 
-The GNU Hurd is distributed in the hope that it will be useful,
-but WITHOUT ANY WARRANTY; without even the implied warranty of
-MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-GNU General Public License for more details.
+   The GNU Hurd is distributed in the hope that it will be useful,
+   but WITHOUT ANY WARRANTY; without even the implied warranty of
+   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+   GNU General Public License for more details.
 
-You should have received a copy of the GNU General Public License
-along with the GNU Hurd; see the file COPYING.  If not, write to
-the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */
+   You should have received a copy of the GNU General Public License
+   along with the GNU Hurd; see the file COPYING.  If not, write to
+   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */
 
 /* Written by Michael I. Bushnell.  */
 
@@ -70,8 +70,8 @@ S_proc_getprivports (struct proc *p,
   if (! check_uid (p, 0))
     return EPERM;
 
-  *hostpriv = master_host_port;
-  *devpriv = master_device_port;
+  *hostpriv = _hurd_host_priv;
+  *devpriv = _hurd_device_master;
   return 0;
 }
 
@@ -419,7 +419,7 @@ S_proc_register_version (pstruct_t server,
 
   /* No need to check SERVER here; we don't use it. */
 
-  if (credential != master_host_port)
+  if (credential != _hurd_host_priv)
     /* Must be privileged to register for uname. */
     return EPERM;
 
